#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import argparse
import csv
import fractions
import json
import logging
import MemoizeDB
import os.path
import sqlite3
import sys
import urllib.error
import urllib.parse
import urllib.request
import xdg.BaseDirectory
import xml.dom.minidom



NAME = 'CurrencyConverter'


WEBSERVICEX_INFO_URL = 'http://www.webservicex.net/CurrencyConvertor.asmx?op=ConversionRate'
WEBSERVICEX_RATES_URL = 'http://www.webservicex.net/CurrencyConvertor.asmx/ConversionRate'

YAHOO_FINANCE_INFO_URL = 'http://finance.yahoo.com/currency-converter/'
# YAHOO_FINANCE_RATES_URL = 'http://finance.yahoo.com/d/quotes.csv?e=.csv&f=c4l1&s={}{}=X' #,GBPUSD=X
YAHOO_FINANCE_RATES_URL = 'http://download.finance.yahoo.com/d/quotes.csv?e=.csv&f=c4l1&s={}'
# s is "<currency a><currency b>=X,<currency c><currency d>=X"

INFO_URL = YAHOO_FINANCE_INFO_URL
RATES_URL = YAHOO_FINANCE_RATES_URL.split('?', 1)[0]



def get_parser():
  parser = argparse.ArgumentParser(
    description='Convert currencies using rates retrieved from {}'.format(RATES_URL),
    epilog='See {} for currency codes.'.format(INFO_URL),
  )
  parser.add_argument('fro', metavar='<fromCurrency>')
  parser.add_argument('to', metavar='<toCurrency>')
  parser.add_argument('amount', metavar='<amount>', nargs='+', action='append')
  parser.add_argument('--ttl', type=int, help='Cached data time-to-live, in seconds.')
  return parser



def urlopen_to_text(f):
  'Convert binary data from urlopen to text data with a generator.'
  for line in f:
    yield line.decode('utf-8')



def get_conversion_rate_from_webservicex(fromCurrency, toCurrency):
  '''
  Retrieve currency conversion rates from

    {}

  Information about recognized currencies can be found on the webpage.
  '''.format(WEBSERVICEX_RATES_URL)
  data = {
    'FromCurrency' : fromCurrency,
    'ToCurrency' : toCurrency,
  }
  data = urllib.parse.urlencode(data).encode('utf-8')
  try:
    with urllib.request.urlopen(WEBSERVICEX_RATES_URL, data) as f:
      return xml.dom.minidom.parse(f).documentElement.firstChild.data,
  except urllib.error.HTTPError as e:
    logging.error('HTTPError: {}\n'.format(e))
    return None,



def get_conversion_rates_from_yahoo_finance(fromTos):
  '''
  Retrieve currency conversion rates from

    {}

  Information about recognized currencies can be found on the webpage.
  '''.format(YAHOO_FINANCE_INFO_URL)
  fromTos = tuple((a.upper(), b.upper()) for a,b in fromTos)
  s = ','.join('{}{}=X'.format(a,b,) for a,b in fromTos)
  url = YAHOO_FINANCE_RATES_URL.format(s)
  try:
    with urllib.request.urlopen(url) as f:
      reader = csv.reader(urlopen_to_text(f), delimiter=',',quotechar='"')
      for (a,b), row in zip(fromTos, reader):
        cur, val = row
        if val == 'N/A':
          yield None
        else:
          val = float(val)
          if cur == a:
            val = 1.0 / val
          yield val
  except urllib.error.HTTPError as e:
    logging.error('HTTPError: {}\n'.format(e))



def get_conversion_rate(fromCurrency, toCurrency):
  for rate in get_conversion_rates(((fromCurrency, toCurrency),)):
    return rate



def get_conversion_rates(fromTos):
  return get_conversion_rates_from_yahoo_finance(fromTos)



class CurrencyConverter():
  '''
  Convert currencies using temporary caching to reduce remote web calls.
  '''

  def __init__(self, ttl=300, conn=None, dpath=None):
    '''
    ttl: time to live, in seconds, of cached rates

    conn: database connection

    dpath: path to an sqlite3 database

    If a connection is given then it will be used, otherwise a connection will
    be created to the database specified by the path. If no path is given then
    a default path will be used.
    '''
    if not conn:
      if not dpath:
        cache_dir = xdg.BaseDirectory.save_cache_path(NAME)
        dpath = os.path.join(cache_dir, 'rates.sqlite3')

      conn = sqlite3.connect(
        dpath,
        detect_types=sqlite3.PARSE_DECLTYPES|sqlite3.PARSE_COLNAMES,
        isolation_level=None
      )

    def glue_func(xys):
      fromTos = tuple(xy.split('-', 1) for xy in xys)
      for xy, rate in zip(xys, get_conversion_rates(fromTos)):
        yield xy, (rate,)

    # Store it as text to avoid rounding errors in the float implementation.
    glue = {
      'rate' : (glue_func, (('rate', 'TEXT'),), ttl)
    }

    self.conn = conn
    self.glue = glue

    self.mdb = MemoizeDB.MemoizeDB(conn, glue)
    self.mdb.db_initialize()
    self.mdb.db_clean()



  def get_rate(self, fromCurrency, toCurrency):
    rate = self.mdb.get_one('rate', "{}-{}".format(fromCurrency,toCurrency))
    if rate is None:
      return float('nan')
    else:
      return fractions.Fraction(rate)


  def convert(self, fromCurrency, toCurrency, amount):
    rate = self.get_rate(fromCurrency, toCurrency)
    try:
      return [rate * x for x in amount]
    except TypeError:
      return rate * amount




def main():
  parser = get_parser()
  pargs = parser.parse_args()

  try:
    cc = CurrencyConverter(ttl=pargs.ttl)
    a = pargs.fro.upper()
    b = pargs.to.upper()
    c = [fractions.Fraction(x) for x in pargs.amount[0]]
    d = ["%0.4f" % (x) for x in c]
    e = ["%0.4f" % (x) for x in cc.convert(a, b, c)]

    l = max(map(len, d))
    r = max(map(len, e))
    hdr_fmt = '{{:<{:d}s}}\t{{:<{:d}s}}'.format(l,r)
    row_fmt = '{{:>{:d}s}}\t{{:>{:d}s}}'.format(l,r)
    print(hdr_fmt.format(a, b))
    print('{}\t{}'.format('-' * l, '-' * r))
    for x,y in zip(d, e):
      print(row_fmt.format(x, y))

  except urllib.error.HTTPError as e:
    sys.exit('error: failed to retrieve conversion rate ({:d}: {})' % (e.code, e.msg))

  except MemoizeDB.MDBError as e:
    sys.exit(str(e))


if __name__ == '__main__':
  try:
    main()
  except (KeyboardInterrupt, BrokenPipeError):
    pass
